@@ -2,21 +2,31 @@
 
 """Stb-tester Python API: Automated GUI Testing for Set-Top Boxes.
 
-Copyright © 2013-2020 Stb-tester.com Ltd. All rights reserved.
+Copyright © 2013-2022 Stb-tester.com Ltd. All rights reserved.
 """
 
+from __future__ import annotations
+
+from contextlib import contextmanager
+from typing import Dict, Generator
+
 import stbt_core  # open source APIs
 
 # We need to import these explicitly so they show up in VSCode Intellisense:
 from stbt_core import (
+    android,
     apply_ocr_corrections,
     as_precondition,
+    BGRDiff,
+    Color,
     ConfigurationError,
     ConfirmMethod,
     crop,
     debug,
     detect_motion,
+    Direction,
     draw_text,
+    find_file,
     for_object_repository,
     Frame,
     FrameDiffer,
@@ -24,12 +34,16 @@ from stbt_core import (
     frames,
     get_config,
     get_frame,
+    GrayscaleDiff,
     Grid,
     Image,
     is_screen_black,
     Keyboard,
     last_keypress,
     load_image,
+    load_mask,
+    Mask,
+    MaskTypes,
     match,
     match_all,
     match_text,
@@ -37,9 +51,9 @@ from stbt_core import (
     MatchParameters,
     MatchResult,
     MatchTimeout,
-    MotionDiff,
     MotionResult,
     MotionTimeout,
+    MultiPress,
     NoVideo,
     ocr,
     OcrEngine,
@@ -53,7 +67,7 @@ from stbt_core import (
     Region,
     save_frame,
     set_global_ocr_corrections,
-    StrictDiff,
+    Size,
     TextMatchResult,
     TransitionStatus,
     UITestError,
@@ -70,13 +84,17 @@ __all__ = stbt_core.__all__ + [
     "audio_chunks",
     "AudioChunk",
     "BadSyncPattern",
+    "color_diff",
     "detect_pages",
+    "find_regions_by_color",
     "find_selection_from_background",
     "FIRST_FRAME_TIME",
     "get_rms_volume",
     "measure_av_sync",
     "play_audio_file",
     "prometheus",
+    "Roku",
+    "segment",
     "stop_job",
     "TEST_PACK_ROOT",
     "VolumeChangeDirection",
@@ -98,7 +116,7 @@ FIRST_FRAME_TIME = None
 TEST_PACK_ROOT = None
 
 
-# pylint: disable=redefined-outer-name,unused-argument
+# pylint: disable=redefined-outer-name,unused-argument,useless-return
 
 
 def _raise_premium(api_name):
@@ -442,7 +460,7 @@ class BadSyncPattern(RuntimeError):
     pass
 
 
-class FindSelectionFromBackgroundResult(object):
+class FindSelectionFromBackgroundResult:
     def __init__(self, matched, region, mask_region, image, frame):
         self.matched = matched
         self.region = region
@@ -484,38 +502,34 @@ def find_selection_from_background(
 
     Unlike `stbt.match`, ``image`` must be the same size as ``frame``.
 
-    :type image: str or `numpy.ndarray`
-    :param image:
+    :param str|stbt.Image image:
       The background to match against. It can be the filename of a PNG file on
-      disk, or a numpy array containing the pixel data in 8-bit BGR format.
+      disk, or an image previously loaded with `stbt.load_image`.
+
       If it has an alpha channel, any transparent pixels are masked out (that
       is, the alpha channel is ANDed with ``mask``). This image must be the
       same size as ``frame``.
 
-    :type max_size: tuple of 2 ints ``(width, height)``
-    :param max_size:
-      The maximum size of the differing region. If the differences between
-      ``image`` and ``frame`` are larger than this in either dimension, the
-      function will return a falsey result.
+    :param stbt.Size max_size:
+      The maximum size ``(width, height)`` of the differing region. If the
+      differences between ``image`` and ``frame`` are larger than this in
+      either dimension, the function will return a falsey result.
 
-    :type min_size: tuple of 2 ints ``(width, height)``
-    :param min_size:
-      The minimum size of the differing region (optional). If the differences
-      between ``image`` and ``frame`` are smaller than this in either
-      dimension, the function will return a falsey result.
+    :param stbt.Size min_size:
+      The minimum size ``(width, height)`` of the differing region (optional).
+      If the differences between ``image`` and ``frame`` are smaller than this
+      in either dimension, the function will return a falsey result.
 
-    :type frame: `stbt.Frame` or `numpy.ndarray`
-    :param frame:
+    :param stbt.Frame frame:
       If this is specified it is used as the video frame to search in;
       otherwise a new frame is grabbed from the device-under-test. This is an
       image in OpenCV format (for example as returned by `stbt.frames` and
       `stbt.get_frame`).
 
-    :type mask: `stbt.Region` or `numpy.ndarray`
-    :param mask:
-      Specifies an area within the image to check. If it's a numpy array
-      (image) it should be a single-channel black-and-white image where black
-      pixels are masked out (ignored).
+    :param str|numpy.ndarray|Mask|Region mask:
+      A `Region` or a mask that specifies which parts of the image to
+      analyse. This accepts anything that can be converted to a Mask using
+      `stbt.load_mask`. See :doc:`masks`.
 
     :type threshold: int
     :param threshold:
@@ -544,7 +558,10 @@ def find_selection_from_background(
         ``find_selection_from_background``.
       * **frame** (`stbt.Frame`) – The video-frame that was analysed.
 
-    Added in v32.
+    ``find_selection_from_background`` was added in v32.
+
+    Changed in v33: ``mask`` accepts anything that can be converted to a Mask
+    using `load_mask` (previously it only accepted a `Region`).
     """
     _raise_premium("find_selection_from_background")
     return FindSelectionFromBackgroundResult(True, Region.ALL, mask, image,
@@ -597,7 +614,7 @@ def detect_pages(frame=None, candidates=None, test_pack_root=""):
 
     Added in v32.
 
-    .. _generator: https://docs.python.org/3.6/tutorial/classes.html#generators
+    .. _generator: https://docs.python.org/3.10/tutorial/classes.html#generators
     """
     _raise_premium("detect_pages")
     return iter([FrameObject()])
@@ -615,6 +632,206 @@ def stop_job(reason=None):
 
     :param str reason: Optional message that will be logged.
 
-    Added in Stb-tester v31.
+    Added in v31.
     """
     _raise_premium("stop_job")
+
+
+def color_diff(frame=None, *, background_color=None, foreground_color=None,
+               threshold=0.05, erode=False):
+    """Calculate euclidean color distance in a perceptually uniform colorspace.
+
+    Calculates the distance of each pixel in ``frame`` against the color
+    specified in ``background_color`` or ``foreground_color``. The output is
+    a binary (black and white) image.
+
+    :param stbt.Frame frame:
+      The video frame to process.
+
+    :param Color background_color:
+      The color to diff against. Output pixels will be white where the color
+      distance is greater than ``threshold``. Use this to remove a background
+      of a particular color.
+
+    :param Color foreground_color:
+      The color to diff against. Output pixels will be white where the color
+      distance is smaller than ``threshold``. Use this to find a foreground
+      feature of a particular color, such as text or the selection/focus.
+
+    :param float threshold:
+      Binarization threshold in the range [0., 1.]. Foreground pixels will be
+      set to white, background pixels to black. A value of 0.01 means a
+      barely-noticeable difference to human perception. To disable binarization
+      set ``threshold=None``; the output will be a grayscale image.
+
+    :param bool erode:
+      Run the thresholded differences through an erosion algorithm to remove
+      noise or small differences (less than 3px).
+
+    :rtype: numpy.ndarray
+    :returns: Binary (black & white) image, or grayscale image if
+      ``threshold=None``.
+
+    Added in v33.
+    """
+    _raise_premium("color_diff")
+    return numpy.array([], dtype=numpy.uint8)
+
+
+def find_regions_by_color(
+        color, *, frame=None, threshold=0.05, erode=False, mask=Region.ALL,
+        min_size=(20, 20), max_size=None):
+    """Find contiguous regions of a particular color.
+
+    :param stbt.Frame frame: The video frame to process.
+    :param Color color: See the ``foreground_color`` parameter of `color_diff`.
+    :param float threshold: See `color_diff`.
+    :param bool erode: See `color_diff`.
+    :param str|numpy.ndarray|Mask|Region mask:
+      A `Region` or a mask that specifies which parts of the image to analyse.
+      This accepts anything that can be converted to a Mask using
+      `stbt.load_mask`. See :doc:`masks`.
+    :param Size min_size:
+      Exclude regions that are smaller than this width or height.
+    :param Size max_size:
+      Exclude regions that are larger than this width or height.
+
+    :rtype: list[stbt.Region]
+    :returns: A list of `stbt.Region` instances.
+
+    Added in v33.
+    """
+    _raise_premium("find_regions_by_color")
+    return []
+
+
+def segment(frame, *, region=Region.ALL, initial_direction=Direction.VERTICAL,
+            steps=1, narrow=True, light_background=False):
+    """Segment (partition) the image into a list of contiguous foreground
+    regions.
+
+    This uses an adaptive threshold algorithm to binarize the image into
+    foreground vs. background pixels. For finer control, you can do the
+    binarization yourself (for example with `stbt.color_diff`) and pass the
+    binarized image to ``segment``.
+
+    :param Frame frame:
+      The video-frame or image to process.
+
+    :param Region region:
+      Only search in this region.
+
+    :param Direction initial_direction:
+      Start scanning in this direction (left-to-right or top-to-bottom).
+
+    :param int steps:
+      Do another segmentation within each region found in the previous step,
+      altering direction between VERTICAL and HORIZONTAL each step. For
+      example, the default values
+      ``steps=1, initial_direction=stbt.Direction.VERTICAL`` will find lines of
+      text; ``steps=2`` will recursively perform segmentation horizontally
+      within each line to find each character in the line (assuming the
+      characters don't overlap due to kerning; overlapping characters will be
+      segmented as a single region).
+
+    :param bool narrow:
+      At the last step, narrow each region in the opposite direction. For
+      example: if you are segmenting lines of text with
+      ``steps=1, initial_direction=stbt.Direction.VERTICAL, narrow=False`` you
+      will get regions with ``y`` & ``bottom`` matching the top & bottom of
+      each line, but with ``x`` & ``right`` set to the left & right edges of
+      the frame (0 and the frame's width, respectively). With ``narrow=True``,
+      each region's ``x`` & ``right`` will be the leftmost / rightmost edge of
+      the line.
+
+    :param bool light_background:
+      By default, the adaptive threshold algorithm assumes foreground pixels
+      are light-coloured and background pixels are dark. Set
+      ``light_background=True`` if foreground pixels are dark (for example
+      black text on a light background).
+
+    :rtype: list[stbt.Region]
+    :returns: A list of `stbt.Region` instances.
+
+    Added in v33.
+    """
+    _raise_premium("segment")
+    return []
+
+
+class Roku:
+    """Helper for interacting with Roku devices over the network.
+
+    This uses Roku's `External Control Protocol`_.
+
+    To find the Roku's IP address and to enable the Roku's network control
+    protocol see <https://stb-tester.com/kb/roku>.
+
+    :param str address: IP address of the Roku.
+
+    Or, use ``Roku.from_config()`` to create an instance using the address
+    configured in the test-pack's configuration files.
+
+    Added in v33.
+
+    .. _External Control Protocol: https://developer.roku.com/en-gb/docs/developer-program/debugging/external-control-api.md
+    """
+    def __init__(self, address: str) -> None:
+        _raise_premium("Roku")
+
+    @staticmethod
+    def from_config() -> "Roku":
+        """Create a ``Roku`` instance from the test-packs's configuration files.
+
+        Expects that the Roku's IP address is specified in
+        ``device_under_test.ip_address``. This configuration belongs
+        in your Stb-tester Node's `Node-specific configuration file`_. For
+        example:
+
+        config/test-farm/stb-tester-00044b80ebeb.conf::
+
+            [device_under_test]
+            device_type = roku
+            ip_address = 192.168.1.7
+
+        :raises ConfigurationError: If Roku IP address not configured.
+
+        .. _Node-specific configuration file: https://stb-tester.com/manual/configuration#node-specific-config
+        """
+        _raise_premium("Roku")
+        return Roku(get_config("device_under_test", "ip_address"))
+
+    @contextmanager
+    def save_logs(self, filename: str = "roku.log") \
+            -> Generator[None, None, None]:
+        """Stream logs from the Roku's `debug console`_ to ``filename``.
+
+        This is a context manager. You can use it as a decorator on your
+        test-case functions::
+
+            import stbt
+            roku = stbt.Roku.from_config()
+
+            @roku.save_logs()
+            def test_launching_my_roku_app():
+                ...
+
+        .. _debug console: https://developer.roku.com/en-gb/docs/developer-program/debugging/debugging-channels.md
+        """
+        _raise_premium("Roku")
+        yield
+
+    def query_apps(self) -> dict[str, str]:
+        """Returns a dict of ``application_id: name`` with all the apps
+        installed on the Roku device.
+        """
+        _raise_premium("Roku")
+        return {}
+
+    def launch_app(self, id_or_name) -> None:
+        """Launches the specified app. Accepts the app's ID or name.
+
+        Use `Roku.query_apps` to find the IDs & names of the apps installed
+        on the Roku.
+        """
+        _raise_premium("Roku")
